@@ -3,24 +3,14 @@
 """
 Copyright (C) 2010 Dariusz Suchojad <dsuch at gefira.pl>
 
-This program is free software: you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation, either version 3 of the License, or
-(at your option) any later version.
-
-This program is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this program.  If not, see <http://www.gnu.org/licenses/>.
+Licensed under LGPLv3, see LICENSE.txt for terms and conditions.
 """
 
 from __future__ import absolute_import, division, print_function, unicode_literals
 
 # stdlib
 import hashlib, itertools, logging, re, ssl, sys, time, traceback, urllib2, uuid
+from base64 import b64encode
 from datetime import datetime
 from urllib import quote_plus, quote
 
@@ -39,6 +29,97 @@ from secwall import wsse
 from secwall.constants import *
 from secwall.core import AuthResult, InvocationContext, SecurityException, SecWallException
 
+def on_ssl_cert(url_config, client_cert, field_prefix, needs_auth_info=True):
+    """ Visit _RequestApp._on_ssl_cert method's docstring.
+    """
+    if client_cert:
+        config_fields = {}
+        for field, value in url_config.items():
+            if field.startswith(field_prefix):
+                config_fields[field.split(field_prefix)[1]] = value
+
+        # There are no fields so the user just wants the connection be
+        # encrypted and the client use client certificate however they're
+        # not interested in the cert's fields - so as long as the CA is
+        # OK (and we know it is because otherwise we wouldn't have gotten
+        # so far), we let the client in.
+        if not config_fields:
+            return True
+        else:
+            subject =  client_cert.get('subject')
+            if not subject:
+                return AuthResult(False, AUTH_CERT_NO_SUBJECT)
+
+            cert_fields = dict((elem[0][0].encode('utf-8'), elem[0][1].encode('utf-8')) for elem in subject)
+            
+            for config_field, config_value in config_fields.items():
+                cert_value = cert_fields.get(config_field)
+                if not cert_value:
+                    return AuthResult(False, AUTH_CERT_NO_VALUE)
+                if cert_value != config_value:
+                    return AuthResult(False, AUTH_CERT_VALUE_MISMATCH)
+            else:
+                auth_result = AuthResult(True, '0')
+                if needs_auth_info:
+                    auth_result.auth_info = dict((quote_plus(k), quote_plus(v)) for k, v in cert_fields.iteritems())
+
+                return auth_result
+
+def on_wsse_pwd(wsse, url_config, data, needs_auth_info=True):
+    """ Visit _RequestApp._on_wsse_pwd method's docstring.
+    """
+    if not data:
+        return AuthResult(False, AUTH_WSSE_NO_DATA)
+
+    request = etree.fromstring(data)
+    try:
+        ok, wsse_username = wsse.validate(request, url_config)
+    except SecurityException, e:
+        return AuthResult(False, AUTH_WSSE_VALIDATION_ERROR, e.description)
+    else:
+        auth_result = AuthResult(True, '0')
+        if needs_auth_info:
+            auth_result.auth_info = {b'wsse-pwd-username': str(wsse_username)}
+
+        return auth_result
+        
+def _on_basic_auth(auth, expected_username, expected_password):
+    """ A low-level call for checking the HTTP Basic Auth credentials.
+    """
+    if not auth:
+        return AUTH_BASIC_NO_AUTH
+
+    prefix = 'Basic '
+    if not auth.startswith(prefix):
+        return AUTH_BASIC_INVALID_PREFIX
+
+    _, auth = auth.split(prefix)
+    auth = auth.strip().decode('base64')
+
+    username, password = auth.split(':', 1)
+
+    if username == expected_username and password == expected_password:
+        return True
+    else:
+        return AUTH_BASIC_USERNAME_OR_PASSWORD_MISMATCH
+
+def on_basic_auth(env, url_config, needs_auth_info=True):
+    """ Visit _RequestApp._on_basic_auth method's docstring.
+    """
+    username = url_config['basic-auth-username']
+    result = _on_basic_auth(env.get('HTTP_AUTHORIZATION', ''), username, url_config['basic-auth-password'])
+    is_success = result is True # Yes, need to check for True
+    
+    auth_result = AuthResult(is_success)
+        
+    if is_success: 
+        if needs_auth_info:
+            auth_result.auth_info = {b'basic-auth-username': quote_plus(username).encode('utf-8')}
+    else:
+        auth_result.code = result
+        
+    return auth_result
+
 class _RequestApp(object):
     """ A WSGI application executed on each request.
     """
@@ -81,12 +162,12 @@ class _RequestApp(object):
 
             # Just in case the user didn't do it, upper-case all the headers
             # to make all the comparisons case-insensitive.
-            url_config['from-client-ignore'][:] = [elem.upper() for elem in url_config.get('from-client-ignore', {})]
+            url_config['from-client-ignore'][:] = map(str.upper, url_config.get('from-client-ignore', {}))
 
         if not seen_default and self.add_default_if_not_found:
             self.urls.append((default_pattern, self.config.default_url_config))
 
-    def __call__(self, env, start_response, client_cert=None):
+    def __call__(self, env, start_response, client_cert=None, client_cert_der=None):
         """ Finds the configuration for the given URL and passes the control on
         to the main request handler. In case no config for the given URL is
         found, a 404 Not Found will be returned to the calling side.
@@ -96,6 +177,9 @@ class _RequestApp(object):
         ctx.auth_result = AuthResult()
         ctx.env = env
 
+        ctx.client_cert = client_cert
+        ctx.client_cert_der = client_cert_der
+
         if self.sign_invocation_id:
             h = hashlib.sha256()
             h.update('{0}:{1}'.format(self.instance_secret, ctx.invocation_id))
@@ -129,6 +213,10 @@ class _RequestApp(object):
     def _on_request(self, ctx, start_response, env, url_config, client_cert, match=None):
         """ Checks security, invokes the backend server, returns the response.
         """
+
+        # If True, we know we're being accessed using SSL.
+        has_ssl = False
+
         # Some quick SSL-related checks first.
         if url_config.get('ssl'):
 
@@ -136,6 +224,9 @@ class _RequestApp(object):
             if env.get('wsgi.url_scheme') != 'https':
                 return self._403(ctx, start_response)
 
+            # OK, we're now sure we're being invoked through SSL.
+            has_ssl = True
+
             # Is the client cert required?
             if url_config.get('ssl-cert') and not client_cert:
                 return self._401(ctx, start_response, self._get_www_auth(url_config, 'ssl-cert'))
@@ -144,20 +235,33 @@ class _RequestApp(object):
         data = data if data else None
         ctx.data = data
 
-        for config_type in self.config.validation_precedence:
-            if config_type in url_config:
-
-                handler = getattr(self, '_on_' + config_type.replace('-', '_'))
-                auth_result = handler(env, url_config, client_cert, data)
-
-                ctx.auth_result = auth_result
-                ctx.config_type = config_type
-
-                if not auth_result:
-                    return self._401(ctx, start_response, self._get_www_auth(url_config, config_type))
-                break
+        # ssl-wrap-only implies 'ssl':True but everyone's free to forget about
+        # setting it so we may wind up ostensibly using SSL yet in reality
+        # we'd be using plain HTTP. That's why there's an additional check
+        # for 'has_ssl' below.
+        if url_config.get('ssl-wrap-only'):
+            if not has_ssl:
+                return self._403(ctx, start_response)
+            else:
+                # There will be no authentication performed and we need to fill
+                # in the information ourselves here so that what goes to logs
+                # doesn't use the ctx's defaults.
+                ctx.auth_result = AuthResult(True, '0')
         else:
-            return self._500(ctx, start_response)
+            for config_type in self.config.validation_precedence:
+                if config_type in url_config:
+
+                    handler = getattr(self, '_on_' + config_type.replace('-', '_'))
+                    auth_result = handler(env, url_config, client_cert, data)
+
+                    ctx.auth_result = auth_result
+                    ctx.config_type = config_type
+
+                    if not auth_result:
+                        return self._401(ctx, start_response, self._get_www_auth(url_config, config_type))
+                    break
+            else:
+                return self._500(ctx, start_response)
 
         rewrite = url_config.get('rewrite')
         if rewrite:
@@ -178,7 +282,7 @@ class _RequestApp(object):
                 req.add_header(name, value)
 
         # Custom headers to be sent to the backend server.
-        for name, value in to_backend_add.items():
+        for name, value in to_backend_add.iteritems():
             req.add_header(name, value)
 
         # Sign and return the invocation ID.
@@ -189,12 +293,12 @@ class _RequestApp(object):
             req.add_header('X-sec-wall-invocation-id-signed', ctx.invocation_id_signed)
 
         if url_config.get('add-auth-info', True):
-            req.add_header('X-sec-wall-auth-info', ctx.auth_result.auth_info)
+            req.add_header('X-sec-wall-auth-info', ctx.auth_result.auth_info.strip())
 
         if url_config.get('sign-auth-info', True):
             h = hashlib.sha256()
             h.update('{0}:{1}:{2}'.format(ctx.invocation_id, self.instance_secret,
-                                          ctx.auth_result.auth_info))
+                                          ctx.auth_result.auth_info.strip()))
             req.add_header('X-sec-wall-auth-info-signed', h.hexdigest())
 
         try:
@@ -228,8 +332,7 @@ class _RequestApp(object):
             for name in ctx.url_config['from-backend-ignore']:
                 headers.pop(name, None)
 
-            for name, value in ctx.url_config['to-client-add'].items():
-                headers[name] = value
+            headers.update(ctx.url_config['to-client-add'])
         else:
             # Special-case the 'Server' header.
             if 'Server' in self.from_backend_ignore:
@@ -267,7 +370,7 @@ class _RequestApp(object):
         header_value = www_auth[config_type]
 
         if config_type in('basic-auth', 'wsse-pwd'):
-            header_value = header_value.format(realm=url_config[config_type + '-' + 'realm'])
+            header_value = header_value.format(realm=url_config[config_type + '-realm'])
         elif config_type == 'digest-auth':
             realm = url_config['digest-auth-realm']
             nonce = uuid.uuid4().hex
@@ -303,84 +406,21 @@ class _RequestApp(object):
         code, status, content_type, description = self.config.internal_server_error
         return self._response(ctx, start_response, code, status, {'Content-Type':content_type}, description)
 
-    def _on_ssl_cert(self, env, url_config, client_cert, data):
+    def _on_ssl_cert(self, env, url_config, client_cert, data, field_prefix='ssl-cert-'):
         """ Validates the client SSL/TLS certificates, its very existence and
         the values of its fields (commonName, organizationName etc.)
         """
-        if client_cert:
-            field_prefix = 'ssl-cert-'
-            config_fields = {}
-            for field, value in url_config.items():
-                if field.startswith(field_prefix):
-                    config_fields[field.split(field_prefix)[1]] = value
-
-            # There are no fields so the user just wants the connection be
-            # encrypted and the client use client certificate however they're
-            # not interested in the cert's fields - so as long as the CA is
-            # OK (and we know it is because otherwise we wouldn't have gotten
-            # so far), we let the client in.
-            if not config_fields:
-                return True
-            else:
-                subject =  client_cert.get('subject')
-                if not subject:
-                    return AuthResult(False, AUTH_CERT_NO_SUBJECT)
-
-                cert_fields = dict(elem[0] for elem in subject)
-
-                for config_field, config_value in config_fields.items():
-                    cert_value = cert_fields.get(config_field)
-                    if not cert_value:
-                        return AuthResult(False, AUTH_CERT_NO_VALUE)
-                    if cert_value != config_value:
-                        return AuthResult(False, AUTH_CERT_VALUE_MISMATCH)
-                else:
-                    auth_result = AuthResult(True, '0')
-                    auth_result.auth_info = dict((quote_plus(k), quote_plus(v)) for k, v in cert_fields.iteritems())
-
-                    return auth_result
+        return on_ssl_cert(url_config, client_cert, field_prefix)
 
     def _on_wsse_pwd(self, env, url_config, unused_client_cert, data):
         """ Uses WS-Security UsernameToken/Password to validate the request.
         """
-        if not data:
-            return AuthResult(False, AUTH_WSSE_NO_DATA)
-
-        request = etree.fromstring(data)
-        try:
-            ok, wsse_username = self.wsse.validate(request, url_config)
-        except SecurityException, e:
-            return AuthResult(False, AUTH_WSSE_VALIDATION_ERROR, e.description)
-        else:
-            auth_result = AuthResult(True, '0')
-            auth_result.auth_info = {b'wsse-pwd-username': str(wsse_username)}
-
-            return auth_result
+        return on_wsse_pwd(self.wsse, url_config, data)
 
     def _on_basic_auth(self, env, url_config, *ignored):
         """ Handles HTTP Basic Authentication.
         """
-        auth = env.get('HTTP_AUTHORIZATION')
-        if not auth:
-            return AuthResult(False, AUTH_BASIC_NO_AUTH)
-
-        prefix = 'Basic '
-        if not auth.startswith(prefix):
-            return AuthResult(False, AUTH_BASIC_INVALID_PREFIX)
-
-        _, auth = auth.split(prefix)
-        auth = auth.strip().decode('base64')
-
-        username, password = auth.split(':', 1)
-
-        if username == url_config['basic-auth-username'] and \
-           password == url_config['basic-auth-password']:
-            auth_result = AuthResult(True, '0')
-            auth_result.auth_info = {b'basic-auth-username': quote_plus(username)}
-
-            return auth_result
-        else:
-            return AuthResult(False, AUTH_BASIC_USERNAME_OR_PASSWORD_MISMATCH)
+        return on_basic_auth(env, url_config)
 
     def _parse_digest_auth(self, auth):
         """ Parses the client's Authorization header and transforms it into
@@ -507,7 +547,7 @@ class _RequestApp(object):
                 return AuthResult(False, AUTH_XPATH_EXPR_MISMATCH)
         else:
             auth_result = AuthResult(True, '0')
-            auth_result.auth_info = [str(e) for e in expressions]
+            auth_result.auth_info = map(str, expressions)
 
             return auth_result
 
@@ -528,9 +568,13 @@ class _RequestHandler(pywsgi.WSGIHandler):
 
         try:
             try:
-                cert = self.socket.getpeercert() if hasattr(self.socket, 'getpeercert') \
-                     else None
-                self.result = self.application(self.environ, self.start_response, cert)
+                if hasattr(self.socket, 'getpeercert'):
+                    cert = self.socket.getpeercert()
+                    cert_der = self.socket.getpeercert(True)
+                else:
+                    cert = None
+                    cert_der = None
+                self.result = self.application(self.environ, self.start_response, cert, cert_der)
                 for data in self.result:
                     if data:
                         self.write(data)
